@@ -5,6 +5,25 @@ pub trait KeccakStateItem<const N: usize>: internal::KeccakItem<N> {}
 // Implement the public trait for all items.
 impl<const N: usize, T: internal::KeccakItem<N>> KeccakStateItem<N> for T {}
 
+#[inline(always)]
+pub(crate) fn get_ij<const N: usize, T: KeccakStateItem<N>>(
+    arr: &[T; 25],
+    i: usize,
+    j: usize,
+) -> T {
+    arr[5 * j + i]
+}
+
+#[inline(always)]
+pub(crate) fn set_ij<const N: usize, T: KeccakStateItem<N>>(
+    arr: &mut [T; 25],
+    i: usize,
+    j: usize,
+    value: T,
+) {
+    arr[5 * j + i] = value;
+}
+
 pub(crate) mod internal {
     /// A trait for multiplexing implementations.
     pub trait KeccakItem<const N: usize>: Clone + Copy {
@@ -15,19 +34,15 @@ pub(crate) mod internal {
         fn and_not_xor(a: Self, b: Self, c: Self) -> Self;
         fn xor_constant(a: Self, c: u64) -> Self;
         fn xor(a: Self, b: Self) -> Self;
-        fn load_block<const RATE: usize>(
-            state: &mut [[Self; 5]; 5],
-            blocks: &[&[u8]; N],
-            start: usize,
-        );
-        fn store_block<const RATE: usize>(state: &[[Self; 5]; 5], blocks: &mut [&mut [u8]; N]);
+        fn load_block<const RATE: usize>(state: &mut [Self; 25], blocks: &[&[u8]; N], start: usize);
+        fn store_block<const RATE: usize>(state: &[Self; 25], blocks: &mut [&mut [u8]; N]);
         fn load_block_full<const RATE: usize>(
-            state: &mut [[Self; 5]; 5],
+            state: &mut [Self; 25],
             blocks: &[[u8; 200]; N],
             start: usize,
         );
-        fn store_block_full<const RATE: usize>(a: &[[Self; 5]; 5], out: &mut [[u8; 200]; N]);
+        fn store_block_full<const RATE: usize>(a: &[Self; 25], out: &mut [[u8; 200]; N]);
         fn split_at_mut_n(a: [&mut [u8]; N], mid: usize) -> ([&mut [u8]; N], [&mut [u8]; N]);
-        fn store<const RATE: usize>(state: &[[Self; 5]; 5], out: [&mut [u8]; N]);
+        fn store<const RATE: usize>(state: &[Self; 25], out: [&mut [u8]; N]);
     }
 }
